(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This is the FunctionsOneLoop.m file, to be used in the FeynMaster environment.
Created by: Jorge Romao
Email: jorge.romao@tecnico.ulisboa.pt
Adapted by: Duarte Fontes
Email: duartefontes@tecnico.ulisboa.pt
Last update: 24.02.2020

Content and usage:

  Coef2[exp, a, b]
  Coef3[exp, a, b, c]
  Coef4[exp, a, b, c, d]
  GetSME[exp]
  SubstituteSME[exp]
  ChangeToB0Unique[exp]
  GetDiv[exp]
  GetFinite[exp]
  OneLoopTID[k,amp]
  Dabcd[a,b,c,d]
  GetPars[k,xamp,verbose]
  IndividualFVkD[k,exp]
  IndividualFVk[k,exp]
  OneLoop4k[k,amp]
  OneLoop4kMod[k,amp]
  MyOneLoop[k,amp]
  MyOneLoopMod[k,amp]
  OneLoopMod[k,amp]
  OneLoopModMon[k,amp]
  GetDenominator[amp]
  GetNumerator[amp]
  TCALoopExp=[exp, Denos, Powk, k1, P, C]
  ChangeToD[exp]
  ChangeTo4[exp]
  TrG5[exp]

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)


(*---------------------------------------------------------------------*)
(* Get coefficients for 2 , 3 and 4 terms *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Functions to extract coefficients up to 4 terms
*)
  
Coef4 =Function[{exp, a, b, c, d}, Coefficient[Coefficient[Coefficient[Coefficient[exp, a], b], c], d]];
Coef3 =Function[{exp, a, b, c}, Coefficient[Coefficient[Coefficient[exp, a], b], c]];
Coef2 =Function[{exp, a, b}, Coefficient[Coefficient[exp, a], b]];
(*---------------------------------------------------------------------*)

  
(*---------------------------------------------------------------------*)
(* GetSME: Get the Standard Matrix Elements*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Gets and prints the Standard Matrix Elements. This is useful to work with amplitudes with spinors
              in the external lines.
*)

  
(* aux1,aux2,exp1,LSME,SMETable *)
  
GetSME=Function[exp,Module[{aux1,aux2,exp1,LSME,SMETable},
exp1=exp//DiracSimplify;  
Clear[StandardMatrixElement];
aux1=FCDiracIsolate[exp1, Head -> StandardMatrixElement] // Expand2[#, {StandardMatrixElement, Pair}] & //  ReplaceRepeated[#,  a_Pair StandardMatrixElement[b_] :>   StandardMatrixElement[a b]] & //   Collect2[#, StandardMatrixElement] &;
aux2 = Select[Variables[aux1], (Head[#] === StandardMatrixElement) &];
(* Print["SME = ",aux2] *)
aux2]];   
(*---------------------------------------------------------------------*)
  

(*---------------------------------------------------------------------*)
(* Substitute the Standard Matrix Elements*)
(* SubstituteSME: Substitute the Standard Matrix Elements*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Substitutes the Standard Matrix Elements in an array ME[j]. This is useful to work with amplitudes with spinors
              in the external lines.
*)

(* SubstituteSME=Function[exp,Module[{aux,exp1,varaux,a,b},
exp1=exp//DiracSimplify;    
Clear[StandardMatrixElement];
aux=FCDiracIsolate[exp1, Head -> StandardMatrixElement] // Expand2[#, {StandardMatrixElement, Pair}] & //  ReplaceRepeated[#,  a_Pair StandardMatrixElement[b_] :>   StandardMatrixElement[a b]] & //   Collect2[#, StandardMatrixElement] &;
varaux = Select[Variables[aux], (Head[#] === StandardMatrixElement) &];        
Set @@ {varaux, Table[ME[j], {j, 1, Length[varaux]}]};
aux]];     *)

SubstituteSME = {##} /. {exp_, what_: {}} :> Module[{aux, exp1, varaux, a, b},
Clear[StandardMatrixElement];
exp1 = exp // DiracSimplify;
aux = FCDiracIsolate[exp1, Head -> StandardMatrixElement] // Expand2[#, {StandardMatrixElement, Pair}] & // ReplaceRepeated[#, a_Pair StandardMatrixElement[b_] :> StandardMatrixElement[a b]] & // Collect2[#, StandardMatrixElement] &;
If[ToString[InputForm[what]] == "{}", 
varaux = Select[Variables[aux], (Head[#] === StandardMatrixElement) &],
varaux = what];
Set @@ {varaux, Table[ME[j], {j, 1, Length[varaux]}]};
aux] &;
  
(*---------------------------------------------------------------------*)  
(* ChangeToB0Unique: Change to B0Unique => B0(0,m^2,m^2) *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Transforms equivalente forms of B0 into an unique form
*)
  
ChangeToB0Unique = 
  Function[exp, Module[{subrule,xx,yy,zz,res},
  subrule={B0[0, 0, xx_] -> B0[0, xx, xx] + 1, 
    B0[yy_, 0, yy_] -> B0[0, yy, yy] + 2, 
    PaVe[0, {0}, {0, xx_}] -> PaVe[0, {0}, {xx, xx}] + 1, 
    PaVe[0, {yy_}, {0, yy_}] -> PaVe[0, {0}, {yy, yy}] + 2,
    A0[zz_] -> zz ( PaVe[0, {0}, {zz, zz}] + 1),
    PaVe[0, {}, {zz_}] -> zz ( PaVe[0, {0}, {zz, zz}] + 1)};
    res=exp /. subrule]];
(*---------------------------------------------------------------------*)



(*---------------------------------------------------------------------*)  
(* GetDiv: Get the divergent Part of Loop Integrals *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 29.01.2020

   Objective: Obtains the coefficient of the divergent part. 
   Notes:
   1) div= 1/epsilon
   2) Warns if Div of higer rank not implemented
*)

GetDiv=Function[exp,Module[{xx,yy,zz,ww,vv,subDiv},
PVList=GetPVList[exp];
CheckIfImplemented[PVList];
If[Implemented,  
subDiv=
{ PaVe[0,{},{xx_}]-> 2 xx div,
 A0[xx_]-> 2 xx div, 
 PaVe[0,{xx_},{yy_,zz_}]-> 2 div,
 B0[xx_,yy_,zz_]-> 2 div,
 PaVe[1,{xx_},{yy_,zz_}]-> - div,
 PaVe[1,{xx_},{yy_,zz_},ww_,vv_]-> - div,
 B1[xx_,yy_,zz_]-> - div,
 PaVe[0,0,{xx_},{yy_,zz_}]-> 1/6 div (3 yy + 3 zz -xx),
 PaVe[0,0,{xx_},{yy_,zz_},ww_,vv_]-> 1/6 div (3 yy + 3 zz -xx),
 PaVe[1,1,{xx_},{yy_,zz_}]-> 2/3 div,
 PaVe[1,1,{xx_},{yy_,zz_},ww_,vv_]-> 2/3 div,
 PaVe[0,0,1,{xx_},{yy_,zz_}]-> div/12  (xx -2 yy -4 zz),
 PaVe[0,0,1,{xx_},{yy_,zz_},ww_,vv_]-> div/12  (xx -2 yy -4 zz),
 PaVe[1,1,1,{xx_},{yy_,zz_}]-> -1/2 div,
 PaVe[1,1,1,{xx_},{yy_,zz_},ww_,vv_]-> -1/2 div,
 PaVe[0,0,{a_,b_,c_},{xx_,yy_,zz_}]-> 1/2 div,
 PaVe[0,0,{a_,b_,c_},{xx_,yy_,zz_},ww_,v_]-> 1/2 div,
 PaVe[0,0,1,{a_,b_,c_},{xx_,yy_,zz_}]-> -1/6 div,
 PaVe[0,0,1,{a_,b_,c_},{xx_,yy_,zz_},ww_,vv_]-> -1/6 div,
 PaVe[0,0,2,{a_,b_,c_},{xx_,yy_,zz_}]-> -1/6 div,
 PaVe[0,0,2,{a_,b_,c_},{xx_,yy_,zz_},ww_,vv_]->-1/6 div,
 PaVe[0,0,0,0,{xx1_,xx2_,xx3_,xx4_,xx5_,xx6_},{yy1_,yy2_,yy3_,yy4_}]-> 1/12 div,
 PaVe[0,0,0,0,{xx1_,xx2_,xx3_,xx4_,xx5_,xx6_},{yy1_,yy2_,yy3_,yy4_},ww_,vv_]-> 1/12 div};
  div Coefficient[exp /.subDiv,div],
Print["Warning: There are PV functions of higher rank for which GetDiv is not yet implemented"]]
]] ;
(*---------------------------------------------------------------------*)

  
WhichPV=Function[exp,Module[{aux0,aux1,aux2,aux3,aux4,aux5,aux6,aux7},
aux0=Expand[exp];
aux1=ToString[aux0];
aux2=StringPosition[aux1, "PaVe["];
If[Length[aux2] !=0,  
aux3=StringTake[aux1, {aux2[[1, 2]], StringLength[aux1]}];
aux4=StringPosition[aux3, "{"];
aux5=StringPosition[aux3, "}"];
aux6=StringTake[aux3, {aux4[[1, 1]], aux5[[1, 1]]}];
nmom=Length[ToExpression[aux6]];
aux7=StringTake[aux3, {2, aux4[[1, 2]] - 3}];
nids=Length[ToExpression[StringJoin["{", aux7, "}"]]];
{nids,nmom},{0,2}]
]];  


CheckIfImplemented=Function[lista,Implemented=True;
Do[If[lista[[i,2]] >6,  Implemented=False];
If[lista[[i,2]] == 6&&lista[[i,1]]>4,  Implemented=False];
If[lista[[i,2]] == 3&&lista[[i,1]]>3,  Implemented=False];
If[lista[[i,2]] == 1&&lista[[i,1]]>3,  Implemented=False];
,{i,1,Length[lista]}]];
  

GetPVList=Function[exp,Module[{aux0,ListL,LenList,Lenaux0,auxFinal0,auxFinal1,aux0i,Lenaux0i,auxi,auxList},
aux0=exp//Expand;  
ListL=False;  
If[Head[exp] === List,LenList=Length[exp];ListL=True;];
If[ListL === False,  
If[Head[aux0] === Plus, Lenaux0 = Length[aux0], Lenaux0 = 1];
If[Lenaux0 == 1, auxFinal0 = {WhichPV[aux0]}, 
    auxFinal0=Table[WhichPV[aux0[[i]]], {i, 1, Lenaux0}]];
,
auxList={};    
Do[aux0i=aux0[[i]];
If[Head[aux0i] === Plus, Lenaux0i = Length[aux0i], Lenaux0i = 1];
If[Lenaux0i == 1, auxi={WhichPV[aux0i]};auxList=Append[auxList,auxi];, 
      auxi=Table[WhichPV[aux0i[[j]]], {j, 1, Lenaux0i}];
      auxList=Append[auxList,auxi];];
,{i,1,LenList}];
    ];
If[ListL,
auxFinal1={};
    Do[auxFinal1=Append[auxFinal1,auxList[[i]][[j]]],{i,1,LenList},{j,1,Length[auxList[[i]]]}];
    auxFinal=auxFinal1;,auxFinal=auxFinal0;
    ];
auxFinal
    ]];
(*---------------------------------------------------------------------*)


(*---------------------------------------------------------------------*)
(* GetFinite: Get the finite term from the divergent Part *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Obtains the finite term coming from the epsilon x 1/epslion terms.
*)

GetFinite=Function[exp, Module[{aux0,aux1,aux3,aux4,aux5,SpinorBackToD},
aux0=exp//FCE;
aux1=GetDiv[aux0];
aux3=aux1 /. {D-> 4-eps,div-> 1/eps};
SpinorBackToD = {Momentum[qq_, 4 - eps] -> Momentum[qq, D]};
aux4=aux3/. SpinorBackToD;
aux5=Coefficient[Normal[Series[aux4,{eps,0,1}]],eps,0];
aux5]];
(*---------------------------------------------------------------------*)

  
(*---------------------------------------------------------------------*)
(* OneLoopTID: OneLoop Function that uses TID *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Implements the TID decomposition method and it is called with the same conventions as OneLoop[k,amp]
*)
  
OneLoopTID = Function[{k, amp}, (amp // TID[#, k, ToPaVe -> True] &)];
(*---------------------------------------------------------------------*)


(*---------------------------------------------------------------------*)  
(* Definition of Dabcd with the conventions of LoopTools *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Define the tensor Dabcd with the conventions of LoopTools for the inplementaion of OneLoop4k
*)
  
LL1aux={SPD[r1,r1],SPD[r1-r2,r1-r2],SPD[r2-r3,r2-r3],SPD[r3,r3],SPD[r2,r2],SPD[r1-r3,r1-r3]}

LL1=Calc[LL1aux]

LL2={m0^2,m1^2,m2^2,m3^2}

ri=Function[a,{FVD[r1,a],FVD[r2,a],FVD[r3,a]}]
pvijk=Function[{i,j,k},PaVe[i,j,k,LL1,LL2]]
pvijkl=Function[{i,j,k,l},PaVe[i,j,k,l,LL1,LL2]]

Dabcd=Function[{a,b,c,d,LL1,LL2}, Term1= (MTD[a,b] MTD[c,d] + MTD[a,c] MTD[b,d] +MTD[a,d] MTD[b,c]) pvijkl[0,0,0,0]; Term2=Sum[(MTD[a,b] ri[c][[i]] ri[d][[j]]+ MTD[b,c] ri[a][[i]] ri[d][[j]]+ MTD[a,c] ri[b][[i]] ri[d][[j]] + MTD[a,d] ri[b][[i]] ri[c][[j]] +  MTD[b,d] ri[a][[i]] ri[c][[j]] + MTD[c,d] ri[a][[i]] ri[b][[j]])  pvijkl[0,0,i,j],{i,1,3},{j,1,3}]; Term3=Sum[ri[a][[i]] ri[b][[j]] ri[c][[k]] ri[d][[l]] pvijkl[i,j,k,l],{i,1,3},{j,1,3},{k,1,3},{l,1,3}];Term1+Term2+Term3];
(*---------------------------------------------------------------------*)

  
(*---------------------------------------------------------------------*)				 
(* GetPars; Get paramenter r1,r2,r3,m0,m1,m2,m3 for the Loop integral with LoopTools conventions *) 
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Obtain the loop paramenters  r1,r2,r3,m0,m1,m2,m3 neeed for the inplementaion of OneLoop4k
   Notes: 
   1) Only Implemented for four denominators

   Internal Parameters: aux1,aux2,aux3,aux4,aux5,aux6,aux7,aux8,qq1,qq2,qq3,qq4,qqq1,qqq2,qqq3,qqq4,mmm0,mmm1,mmm2,mmm3,Den1,Den2,Den3,Den4,Den1Comma,Den2Comma,Den3Comma,Den4Comma

*)
				 
GetPars=Function[{k,xamp,verbose},Module[{aux1,aux2,aux3,aux4,aux5,aux6,aux7,aux8,qq1,qq2,qq3,qq4,qqq1,qqq2,qqq3,qqq4,mmm0,mmm1,mmm2,mmm3,Den1,Den2,Den3,Den4,Den1Comma,Den2Comma,Den3Comma,Den4Comma},
aux1=InputForm[Calc[xamp]//FCE//Simplify];			       
aux5=GetDenominator[aux1];
If[Length[aux5]!=4,
Error="True",
Error="False";
aux6=aux5//FCI;
aux7=ToString[aux6];
aux8 = StringPosition[aux7, "PropagatorDenominator["];
len=StringLength[aux7];

Den1=StringTake[aux7, {aux8[[1]][[2]] + 1, aux8[[2]][[1]] - 4}];
Den2=StringTake[aux7, {aux8[[2]][[2]] + 1, aux8[[3]][[1]] - 4}];
Den3=StringTake[aux7, {aux8[[3]][[2]] + 1, aux8[[4]][[1]] - 4}];
Den4=StringTake[aux7, {aux8[[4]][[2]] + 1, len - 2}];

Den1Comma=StringPosition[Den1, ","];
Den2Comma=StringPosition[Den2, ","];
Den3Comma=StringPosition[Den3, ","];
Den4Comma=StringPosition[Den4, ","];
  
qqq1=ToExpression[StringTake[Den1,{1, Den1Comma[[Length[Den1Comma]]][[1]]-1}]];
qqq2=ToExpression[StringTake[Den2,{1, Den2Comma[[Length[Den2Comma]]][[1]]-1}]];
qqq3=ToExpression[StringTake[Den3,{1, Den3Comma[[Length[Den3Comma]]][[1]]-1}]];
qqq4=ToExpression[StringTake[Den4,{1, Den4Comma[[Length[Den4Comma]]][[1]]-1}]];

mmm0 = ToExpression[StringTake[Den1, {Den1Comma[[Length[Den1Comma]]][[1]] + 1, StringLength[Den1]}]];
mmm1 = ToExpression[StringTake[Den2, {Den2Comma[[Length[Den2Comma]]][[1]] + 1, StringLength[Den2]}]];
mmm2 = ToExpression[StringTake[Den3, {Den3Comma[[Length[Den3Comma]]][[1]] + 1, StringLength[Den3]}]];
mmm3 = ToExpression[StringTake[Den4, {Den4Comma[[Length[Den4Comma]]][[1]] + 1, StringLength[Den4]}]];
  
	       
If[(ToString[(qqq1-Momentum[k,D])]  == "0")|| (ToString[(qqq1+Momentum[k,D])]  == "0"),
   qq1=qqq1;
   qq2=qqq2;
   qq3=qqq3;
   qq4=qqq4;
   mm0=mmm0;
   mm1=mmm1;
   mm2=mmm2;
   mm3=mmm3;
   ];

If[(ToString[(qqq2-Momentum[k,D])]  == "0")|| (ToString[(qqq2+Momentum[k,D])]  == "0"),
   qq1=qqq2;
   qq2=qqq1;
   qq3=qqq3;
   qq4=qqq4;
   mm0=mmm1;
   mm1=mmm0;
   mm2=mmm2;
   mm3=mmm3;
   ];

If[(ToString[(qqq3-Momentum[k,D])]  == "0")|| (ToString[(qqq3+Momentum[k,D])]  == "0"),
   qq1=qqq3;
   qq2=qqq1;
   qq3=qqq2;
   qq4=qqq4;
   mm0=mmm2;
   mm1=mmm0;
   mm2=mmm1;
   mm3=mmm3;
   ];


If[(ToString[(qqq4-Momentum[k,D])]  == "0")|| (ToString[(qqq4+Momentum[k,D])]  == "0"),
   qq1=qqq4;
   qq2=qqq1;
   qq3=qqq2;
   qq4=qqq3;
   mm0=mmm3;
   mm1=mmm0;
   mm2=mmm1;
   mm3=mmm2;
   ];


	       
If[Coefficient[qq2-qq1,Momentum[k,D]]!=0,qq2=-qq2];
If[Coefficient[qq3-qq1,Momentum[k,D]]!=0,qq3=-qq3];
If[Coefficient[qq4-qq1,Momentum[k,D]]!=0,qq4=-qq4];	       	       

    
rr1=qq2-qq1;
rr2=qq3-qq1;
rr3=qq4-qq1;	       
If[verbose,Print["r1=",rr1,", r2=",rr2,", r3=",rr3,", m0=",mm0,", m1=",mm1,", m2=",mm2,", m3=",mm3]]
	       
]]]
(*---------------------------------------------------------------------*) 

  
(*---------------------------------------------------------------------*) 
(* IndividualFVkD: Transform Numerator into individual momenta in Dimension D*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Transform Numerator into individual momenta in Dimension D=4
   Notes: 
   1) Present index= i59
   2) Correct up to k^6 
*)


  
IndividualFVkD=Function[{k,exp}, Module[{expAux,expAux0,expAux1,expAux2,expAux3,expAux4,expAux5,expAux6,expAux7,expAux8,expAux9,expAux10,expAux11,expAux12,expAux13,expAux14,expAux15,expAux16,MatAux1,MatAux2},

expAux=exp//FCI//Expand;

expAux0a=expAux/. Pair[Momentum[k,D], Momentum[k,D]]^4 ->MTD[i54, i55] Pair[LorentzIndex[i54,D], Momentum[k,D]] Pair[LorentzIndex[i55,D],Momentum[k,D]] Pair[Momentum[k,D], Momentum[k,D]]^3;

expAux0b=expAux0a/. Pair[Momentum[k,D], Momentum[k,D]]^3 ->MTD[i1, i2] Pair[LorentzIndex[i1,D], Momentum[k,D]] Pair[LorentzIndex[i2,D],Momentum[k,D]] Pair[Momentum[k,D], Momentum[k,D]]^2;

expAux1=expAux0b/. Pair[Momentum[k,D], Momentum[k,D]]^2 -> 
  MTD[i3, i4] Pair[LorentzIndex[i3,D], Momentum[k,D]] Pair[LorentzIndex[i4,D],
 Momentum[k,D]] Pair[Momentum[k,D], Momentum[k,D]];
					
expAux2 = expAux1 /. Pair[Momentum[k,D], Momentum[k,D]] -> MTD[i5, i6] Pair[LorentzIndex[i5,D], Momentum[k,D]] Pair[LorentzIndex[i6,D], Momentum[k,D]];
					
expAux3a = expAux2 /. Pair[Momentum[k,D], Momentum[q_,D]]^4 ->Pair[LorentzIndex[i46,D], Momentum[k,D]] Pair[LorentzIndex[i46,D], Momentum[q,D]] Pair[LorentzIndex[i47,D], Momentum[k,D]] Pair[LorentzIndex[i47,D],  Momentum[q,D]] Pair[LorentzIndex[i48,D], Momentum[k,D]] Pair[LorentzIndex[i48,D],  Momentum[q,D]] Pair[LorentzIndex[i49,D], Momentum[k,D]] Pair[LorentzIndex[i49,D],  Momentum[q,D]];

expAux3b = expAux3a /. Pair[Momentum[k,D], Momentum[p_,D]]^2 Pair[Momentum[k,D], Momentum[q_,D]]^2 ->Pair[LorentzIndex[i50,D], Momentum[k,D]] Pair[LorentzIndex[i50,D], Momentum[q,D]] Pair[LorentzIndex[i51,D], Momentum[k,D]] Pair[LorentzIndex[i51,D],  Momentum[q,D]] Pair[LorentzIndex[i52,D], Momentum[k,D]] Pair[LorentzIndex[i52,D],  Momentum[p,D]] Pair[LorentzIndex[i53,D], Momentum[k,D]] Pair[LorentzIndex[i53,D],  Momentum[p,D]];
			
expAux3c = expAux3b /. Pair[Momentum[k,D], Momentum[q_,D]]^3 ->Pair[LorentzIndex[i43,D], Momentum[k,D]] Pair[LorentzIndex[i43,D], Momentum[q,D]] Pair[LorentzIndex[i44,D], Momentum[k,D]] Pair[LorentzIndex[i44,D],  Momentum[q,D]] Pair[LorentzIndex[i45,D], Momentum[k,D]] Pair[LorentzIndex[i45,D],  Momentum[q,D]];

expAux3d = expAux3c /. Pair[Momentum[k,D], Momentum[q_,D]]^2 ->Pair[LorentzIndex[i7,D], Momentum[k,D]] Pair[LorentzIndex[i7,D], Momentum[q,D]] Pair[LorentzIndex[i8,D], Momentum[k,D]] Pair[LorentzIndex[i8,D],  Momentum[q,D]];

expAux3e = expAux3d /. Pair[Momentum[k,D], Momentum[q_,D]] Pair[Momentum[k,D], Momentum[p_,D]] ->Pair[LorentzIndex[i41,D], Momentum[k,D]] Pair[LorentzIndex[i41,D], Momentum[q,D]] Pair[LorentzIndex[i42,D], Momentum[k,D]] Pair[LorentzIndex[i42,D],  Momentum[p,D]];
					
expAux4 = expAux3e /. Pair[Momentum[k,D], Momentum[q_,D]]-> Pair[LorentzIndex[i9,D], Momentum[k,D]] Pair[LorentzIndex[i9,D], Momentum[q,D]];

expAux5=expAux4 /.   MatAux1_ . DiracGamma[Momentum[k,D],D] .	DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] . MatAux2_ ->   MatAux1 . DiracGamma[LorentzIndex[i10,D],D] .	DiracGamma[LorentzIndex[i11,D],D] . DiracGamma[LorentzIndex[i12,D],D] . MatAux2 Pair[LorentzIndex[i10,D], Momentum[k,D]] Pair[LorentzIndex[i11,D], Momentum[k,D]] Pair[LorentzIndex[i12,D], Momentum[k,D]];

expAux6=expAux5 /.    DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] . MatAux2_ ->   DiracGamma[LorentzIndex[i13,D],D] . DiracGamma[LorentzIndex[i14,D],D] . DiracGamma[LorentzIndex[i15,D],D] . MatAux2 Pair[LorentzIndex[i13,D], Momentum[k,D]] Pair[LorentzIndex[i14,D], Momentum[k,D]] Pair[LorentzIndex[i15,D], Momentum[k,D]];					

expAux7=expAux6 /.  MatAux1_ . DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D]  ->  MatAux1.  DiracGamma[LorentzIndex[i16,D],D] . DiracGamma[LorentzIndex[i17,D],D] . DiracGamma[LorentzIndex[i18,D],D]  Pair[LorentzIndex[i16,D], Momentum[k,D]] Pair[LorentzIndex[i17,D], Momentum[k,D]] Pair[LorentzIndex[i18,D], Momentum[k,D]];					
					
expAux8=expAux7 /.   DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D]  ->  DiracGamma[LorentzIndex[i19,D],D] .	DiracGamma[LorentzIndex[i20,D],D] . DiracGamma[LorentzIndex[i21,D],D]  Pair[LorentzIndex[i19,D], Momentum[k,D]] Pair[LorentzIndex[i20,D], Momentum[k,D]] Pair[LorentzIndex[i21,D], Momentum[k,D]];					

expAux9=expAux8 /.   MatAux1_ . DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] . MatAux2_ ->   MatAux1 . DiracGamma[LorentzIndex[i21,D],D] .	DiracGamma[LorentzIndex[i22,D],D]  . MatAux2 Pair[LorentzIndex[i21,D], Momentum[k,D]] Pair[LorentzIndex[i22,D], Momentum[k,D]] ;
							
expAux10=expAux9 /.   MatAux1_ . DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] ->   MatAux1 . DiracGamma[LorentzIndex[i23,D],D] .	DiracGamma[LorentzIndex[i24,D],D]   Pair[LorentzIndex[i23,D], Momentum[k,D]] Pair[LorentzIndex[i24,D], Momentum[k,D]] ;
							
expAux11=expAux10 /.    DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D] . MatAux2_ ->    DiracGamma[LorentzIndex[i25,D],D] . DiracGamma[LorentzIndex[i26,D],D]  . MatAux2 Pair[LorentzIndex[i25,D], Momentum[k,D]] Pair[LorentzIndex[i26,D], Momentum[k,D]] ;
							
expAux12=expAux11 /.   DiracGamma[Momentum[k,D],D] . DiracGamma[Momentum[k,D],D]  ->   DiracGamma[LorentzIndex[i27,D],D] . DiracGamma[LorentzIndex[i28,D],D]  Pair[LorentzIndex[i27,D], Momentum[k,D]] Pair[LorentzIndex[i28,D], Momentum[k,D]] ;

expAux13=expAux12 /. MatAux1_ . DiracGamma[Momentum[k,D],D] . MatAux2_-> MatAux1 . DiracGamma[LorentzIndex[i28,D],D] . MatAux2  Pair[LorentzIndex[i28,D], Momentum[k,D]];

expAux14=expAux13 /. MatAux1_ . DiracGamma[Momentum[k,D],D] -> MatAux1 . DiracGamma[LorentzIndex[i29,D],D]  Pair[LorentzIndex[i29,D], Momentum[k,D]];

expAux15=expAux14 /.  DiracGamma[Momentum[k,D],D] . MatAux2_->  DiracGamma[LorentzIndex[i30,D],D] . MatAux2  Pair[LorentzIndex[i30,D], Momentum[k,D]];

expAux16=expAux15 /.  DiracGamma[Momentum[k,D],D] ->  DiracGamma[LorentzIndex[i31,D],D]  Pair[LorentzIndex[i31,D], Momentum[k,D]];

(* 
Next four entries are because of Levi-Civita Eps. Notice that it is in D=4. 
Hence we had to modify the derD definition for this case.
This only works when there is only one Momentum[k] in the Eps 
20.09.2018 jcr
*)
  
expAux17=expAux16 /. Eps[Momentum[k],li1_,li2_,li3_]-> Eps[LorentzIndex[i56],li1,li2,li3]  Pair[LorentzIndex[i56],Momentum[k]];

expAux18=expAux17 /. Eps[li1_,Momentum[k],li2_,li3_]-> Eps[li1,LorentzIndex[i57],li2,li3]  Pair[LorentzIndex[i57],Momentum[k]];

expAux19=expAux18 /. Eps[li1_,li2_,Momentum[k],li3_]-> Eps[li1,li2,LorentzIndex[i58],li3]  Pair[LorentzIndex[i58],Momentum[k]];

expAux20=expAux19 /. Eps[li1_,li2_,li3_,Momentum[k]]-> Eps[li1,li2,li3,LorentzIndex[i59]]  Pair[LorentzIndex[i59],Momentum[k]];  expAux20]];
(*---------------------------------------------------------------------*) 	

  
(*---------------------------------------------------------------------*) 
(* Definitions of derivatives with respect to 4-momenta in D dimensions*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Defines the derivatives of the expression with respect to 4-momenta in Dimension D.
   Notes:
   1) Modified for epsilon terms in 01.08.2019 
*)
  
derD[Pair[LorentzIndex[i_,D], Momentum[k_,D]], 
      Pair[LorentzIndex[j_,D], Momentum[k_,D]]] := MTD[i, j]

derD[Pair[LorentzIndex[i_], Momentum[k_]], 
Pair[LorentzIndex[j_,D], Momentum[k_,D]]] := MT[i, j]

derD[a_  b_, Pair[LorentzIndex[i_,D], Momentum[k_,D]]] := 
derD[a, Pair[LorentzIndex[i,D], Momentum[k,D]]] b + 
a derD[b, Pair[LorentzIndex[i,D], Momentum[k,D]]]

derD[a_ + b_, Pair[LorentzIndex[i_,D], Momentum[k_,D]]] := 
derD[a, Pair[LorentzIndex[i,D], Momentum[k,D]]]  + 
derD[b, Pair[LorentzIndex[i,D], Momentum[k,D]]]

		 
derD[c_, Pair[LorentzIndex[i_,D], Momentum[k_,D]]] := 
0 /; FreeQ[c, Pair[LorentzIndex[i,D], Momentum[k,D]]]
(*---------------------------------------------------------------------*) 


(*---------------------------------------------------------------------*) 		 
(* IndividualFVk: Transform Numerator into individual momenta in Dimension D=4 *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Transform Numerator into individual momenta in Dimension D=4
   Notes: 
   1) Present index= i59
   2) Correct up to k^6 
   3) Note completely checked in D=4, but this not important as in the code is used the Dimension D version
*)
  
IndividualFVk=Function[{k,exp}, Module[{expAux,expAux0,expAux1,expAux2,expAux3,expAux4,expAux5,expAux6,expAux7,expAux8,expAux9,expAux10,expAux11,expAux12,expAux13,expAux14,expAux15,expAux16,MatAux1,MatAux2},

expAux=exp//FCI//Expand;

				       
expAux0a=expAux/. Pair[Momentum[k], Momentum[k]]^4 ->MT[i54, i55] Pair[LorentzIndex[i54], Momentum[k]] Pair[LorentzIndex[i55],Momentum[k]] Pair[Momentum[k], Momentum[k]]^3;

expAux0b=expAux0a/. Pair[Momentum[k], Momentum[k]]^3 ->MT[i1, i2] Pair[LorentzIndex[i1], Momentum[k]] Pair[LorentzIndex[i2],Momentum[k]] Pair[Momentum[k], Momentum[k]]^2;

expAux1=expAux0b/. Pair[Momentum[k], Momentum[k]]^2 -> 
  MT[i3, i4] Pair[LorentzIndex[i3], Momentum[k]] Pair[LorentzIndex[i4],
 Momentum[k]] Pair[Momentum[k], Momentum[k]];
					
expAux2 = expAux1 /. Pair[Momentum[k], Momentum[k]] -> MT[i5, i6] Pair[LorentzIndex[i5], Momentum[k]] Pair[LorentzIndex[i6], Momentum[k]];


expAux3a1 = expAux2 /. Pair[Momentum[k], Momentum[q_]]^6 ->Pair[LorentzIndex[i60], Momentum[k]] Pair[LorentzIndex[i60], Momentum[q]] Pair[LorentzIndex[i61], Momentum[k]] Pair[LorentzIndex[i61],  Momentum[q]] Pair[LorentzIndex[i62], Momentum[k]] Pair[LorentzIndex[i62],  Momentum[q]] Pair[LorentzIndex[i63], Momentum[k]] Pair[LorentzIndex[i63],  Momentum[q]] Pair[LorentzIndex[i64], Momentum[k]] Pair[LorentzIndex[i64],  Momentum[q]] Pair[LorentzIndex[i65], Momentum[k]] Pair[LorentzIndex[i65],  Momentum[q]];

expAux3a2 = expAux3a1 /. Pair[Momentum[k], Momentum[q_]]^3 Pair[Momentum[k], Momentum[p_]]^3 ->Pair[LorentzIndex[i71], Momentum[k]] Pair[LorentzIndex[i71], Momentum[q]] Pair[LorentzIndex[i72], Momentum[k]] Pair[LorentzIndex[i72],  Momentum[q]] Pair[LorentzIndex[i73], Momentum[k]] Pair[LorentzIndex[i73],  Momentum[q]] Pair[LorentzIndex[i74], Momentum[k]] Pair[LorentzIndex[i74],  Momentum[p]] Pair[LorentzIndex[i75], Momentum[k]] Pair[LorentzIndex[i75],  Momentum[p]] Pair[LorentzIndex[i76], Momentum[k]] Pair[LorentzIndex[i76],  Momentum[p]];
				       

expAux3a3 = expAux3a2 /. Pair[Momentum[k], Momentum[q_]]^5 ->Pair[LorentzIndex[i66], Momentum[k]] Pair[LorentzIndex[i66], Momentum[q]] Pair[LorentzIndex[i67], Momentum[k]] Pair[LorentzIndex[i67],  Momentum[q]] Pair[LorentzIndex[i68], Momentum[k]] Pair[LorentzIndex[i68],  Momentum[q]] Pair[LorentzIndex[i69], Momentum[k]] Pair[LorentzIndex[i69],  Momentum[q]] Pair[LorentzIndex[i70], Momentum[k]] Pair[LorentzIndex[i70],  Momentum[q]];
		       
				       
expAux3a = expAux3a3 /. Pair[Momentum[k], Momentum[q_]]^4 ->Pair[LorentzIndex[i46], Momentum[k]] Pair[LorentzIndex[i46], Momentum[q]] Pair[LorentzIndex[i47], Momentum[k]] Pair[LorentzIndex[i47],  Momentum[q]] Pair[LorentzIndex[i48], Momentum[k]] Pair[LorentzIndex[i48],  Momentum[q]] Pair[LorentzIndex[i49], Momentum[k]] Pair[LorentzIndex[i49],  Momentum[q]];

expAux3b = expAux3a /. Pair[Momentum[k], Momentum[p_]]^2 Pair[Momentum[k], Momentum[q_]]^2 ->Pair[LorentzIndex[i50], Momentum[k]] Pair[LorentzIndex[i50], Momentum[q]] Pair[LorentzIndex[i51], Momentum[k]] Pair[LorentzIndex[i51],  Momentum[q]] Pair[LorentzIndex[i52], Momentum[k]] Pair[LorentzIndex[i52],  Momentum[p]] Pair[LorentzIndex[i53], Momentum[k]] Pair[LorentzIndex[i53],  Momentum[p]];
			
expAux3c = expAux3b /. Pair[Momentum[k], Momentum[q_]]^3 ->Pair[LorentzIndex[i43], Momentum[k]] Pair[LorentzIndex[i43], Momentum[q]] Pair[LorentzIndex[i44], Momentum[k]] Pair[LorentzIndex[i44],  Momentum[q]] Pair[LorentzIndex[i45], Momentum[k]] Pair[LorentzIndex[i45],  Momentum[q]];

expAux3d = expAux3c /. Pair[Momentum[k], Momentum[q_]]^2 ->Pair[LorentzIndex[i7], Momentum[k]] Pair[LorentzIndex[i7], Momentum[q]] Pair[LorentzIndex[i8], Momentum[k]] Pair[LorentzIndex[i8],  Momentum[q]];

expAux3e = expAux3d /. Pair[Momentum[k], Momentum[q_]] Pair[Momentum[k], Momentum[p_]] ->Pair[LorentzIndex[i41], Momentum[k]] Pair[LorentzIndex[i41], Momentum[q]] Pair[LorentzIndex[i42], Momentum[k]] Pair[LorentzIndex[i42],  Momentum[p]];
					
expAux4 = expAux3e /. Pair[Momentum[k], Momentum[q_]]-> Pair[LorentzIndex[i9], Momentum[k]] Pair[LorentzIndex[i9], Momentum[q]];

expAux5=expAux4 /.   MatAux1_ . DiracGamma[Momentum[k]] .	DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] . MatAux2_ ->   MatAux1 . DiracGamma[LorentzIndex[i10]] .	DiracGamma[LorentzIndex[i11]] . DiracGamma[LorentzIndex[i12]] . MatAux2 Pair[LorentzIndex[i10], Momentum[k]] Pair[LorentzIndex[i11], Momentum[k]] Pair[LorentzIndex[i12], Momentum[k]];

expAux6=expAux5 /.    DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] . MatAux2_ ->   DiracGamma[LorentzIndex[i13]] . DiracGamma[LorentzIndex[i14]] . DiracGamma[LorentzIndex[i15]] . MatAux2 Pair[LorentzIndex[i13], Momentum[k]] Pair[LorentzIndex[i14], Momentum[k]] Pair[LorentzIndex[i15], Momentum[k]];					

expAux7=expAux6 /.  MatAux1_ . DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]]  ->  MatAux1.  DiracGamma[LorentzIndex[i16]] . DiracGamma[LorentzIndex[i17]] . DiracGamma[LorentzIndex[i18]]  Pair[LorentzIndex[i16], Momentum[k]] Pair[LorentzIndex[i17], Momentum[k]] Pair[LorentzIndex[i18], Momentum[k]];					
					
expAux8=expAux7 /.   DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]]  ->  DiracGamma[LorentzIndex[i19]] .	DiracGamma[LorentzIndex[i20]] . DiracGamma[LorentzIndex[i21]]  Pair[LorentzIndex[i19], Momentum[k]] Pair[LorentzIndex[i20], Momentum[k]] Pair[LorentzIndex[i21], Momentum[k]];					

expAux9=expAux8 /.   MatAux1_ . DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] . MatAux2_ ->   MatAux1 . DiracGamma[LorentzIndex[i21]] .	DiracGamma[LorentzIndex[i22]]  . MatAux2 Pair[LorentzIndex[i21], Momentum[k]] Pair[LorentzIndex[i22], Momentum[k]] ;
							
expAux10=expAux9 /.   MatAux1_ . DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] ->   MatAux1 . DiracGamma[LorentzIndex[i23]] .	DiracGamma[LorentzIndex[i24]]   Pair[LorentzIndex[i23], Momentum[k]] Pair[LorentzIndex[i24], Momentum[k]] ;
							
expAux11=expAux10 /.    DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]] . MatAux2_ ->    DiracGamma[LorentzIndex[i25]] . DiracGamma[LorentzIndex[i26]]  . MatAux2 Pair[LorentzIndex[i25], Momentum[k]] Pair[LorentzIndex[i26], Momentum[k]] ;
							
expAux12=expAux11 /.   DiracGamma[Momentum[k]] . DiracGamma[Momentum[k]]  ->   DiracGamma[LorentzIndex[i27]] . DiracGamma[LorentzIndex[i28]]  Pair[LorentzIndex[i27], Momentum[k]] Pair[LorentzIndex[i28], Momentum[k]] ;

expAux13=expAux12 /. MatAux1_ . DiracGamma[Momentum[k]] . MatAux2_-> MatAux1 . DiracGamma[LorentzIndex[i28]] . MatAux2  Pair[LorentzIndex[i28], Momentum[k]];

expAux14=expAux13 /. MatAux1_ . DiracGamma[Momentum[k]] -> MatAux1 . DiracGamma[LorentzIndex[i29]]  Pair[LorentzIndex[i29], Momentum[k]];

expAux15=expAux14 /.  DiracGamma[Momentum[k]] . MatAux2_->  DiracGamma[LorentzIndex[i30]] . MatAux2  Pair[LorentzIndex[i30], Momentum[k]];

expAux16=expAux15 /.  DiracGamma[Momentum[k]] ->  DiracGamma[LorentzIndex[i31]]  Pair[LorentzIndex[i31], Momentum[k]];

(* 
Next four entries are because of Levi-Civita Eps. Notice that it is in D=4. 
Hence we had to modify the derD definition for this case.
This only works when there is only one Momentum[k] in the Eps 
20.09.2018 jcr
*)
  
expAux17=expAux16 /. Eps[Momentum[k],li1_,li2_,li3_]-> Eps[LorentzIndex[i56],li1,li2,li3]  Pair[LorentzIndex[i56],Momentum[k]];

expAux18=expAux17 /. Eps[li1_,Momentum[k],li2_,li3_]-> Eps[li1,LorentzIndex[i57],li2,li3]  Pair[LorentzIndex[i57],Momentum[k]];

expAux19=expAux18 /. Eps[li1_,li2_,Momentum[k],li3_]-> Eps[li1,li2,LorentzIndex[i58],li3]  Pair[LorentzIndex[i58],Momentum[k]];

expAux20=expAux19 /. Eps[li1_,li2_,li3_,Momentum[k]]-> Eps[li1,li2,li3,LorentzIndex[i59]]  Pair[LorentzIndex[i59],Momentum[k]];  expAux20]];
(*---------------------------------------------------------------------*)                	
                 
	 
(*---------------------------------------------------------------------*)                
(* Definitions of derivatives with respect to 4-momenta in D=4*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Defines the derivatives of the expression with respect to 4-momenta in D=4.
*)
   
der[Pair[LorentzIndex[i_], Momentum[k_]], 
  Pair[LorentzIndex[j_], Momentum[k_]]] := MT[i, j]

der[a_  b_, Pair[LorentzIndex[i_], Momentum[k_]]] := 
 der[a, Pair[LorentzIndex[i], Momentum[k]]] b + 
  a der[b, Pair[LorentzIndex[i], Momentum[k]]]

der[a_ + b_, Pair[LorentzIndex[i_], Momentum[k_]]] := 
 der[a, Pair[LorentzIndex[i], Momentum[k]]]  + 
   der[b, Pair[LorentzIndex[i], Momentum[k]]]

der[(DiracGamma[LorentzIndex[i_]] Pair[LorentzIndex[j_], 
     Momentum[k_]]), Pair[LorentzIndex[l_], Momentum[k_]]] :=
DiracGamma[LorentzIndex[i]] MT[j,l]

der[ A_ . (DiracGamma[LorentzIndex[i_]] Pair[LorentzIndex[j_], 
     Momentum[k_]]), Pair[LorentzIndex[l_], Momentum[k_]]] :=
A . DiracGamma[LorentzIndex[i]] MT[j,l]

der[c_, Pair[LorentzIndex[i_], Momentum[k_]]] := 
 0 /; FreeQ[c, Pair[LorentzIndex[i], Momentum[k]]]

(*---------------------------------------------------------------------*)

		
(*---------------------------------------------------------------------*)
(* OneLoop4k: Function to calculate OneLoop Integrals with 4 momenta*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Extends the OneLoop function of FC to the case of 
   4 momenta in the numerator.
   
   Internal variables: aux1,aux2,amp3k,amp4k,aux3,aux4,aux5,subPars

*)
		 
OneLoop4k=Function[{k,amp},Module[{aux1,aux2,amp3k,amp4k,aux3,aux4,aux5,subPars},
aux1=Calc[amp]//FCI;
aux2=aux1 /. Momentum[k, D] -> x Momentum[k, D];
amp4k=Coefficient[aux2, x, 4] /. x -> 1 // FCE;			  
amp3k=Calc[(aux1 - amp4k)]//Simplify;
GetPars[k,amp4k,False];
If[Error=="False",  
If[ToString[amp3k] != "0", Res3k=OneLoop[k,amp3k],Res3k=0;];
Num4k=GetNumerator[amp4k];
subPars={Momentum[r1,D] -> Momentum[rr1,D], Momentum[r2,D]->Momentum[rr2,D],Momentum[r3,D]->Momentum[rr3,D] ,m0->mm0,m1->mm1,m2->mm2,m3->mm3} ;
aux3=IndividualFVkD[k,Num4k] ;
aux4 = 
  derD[derD[
      derD[derD[aux3, Pair[LorentzIndex[j1, D], Momentum[k, D]]], 
      Pair[LorentzIndex[j2, D], Momentum[k, D]]], 
      Pair[LorentzIndex[j3, D], Momentum[k, D]]], 
      Pair[LorentzIndex[j4, D], Momentum[k, D]]]/4! // Expand;
aux5 = Calc[aux4 Dabcd[j1, j2, j3, j4, LL1, LL2]];
Res4kD=aux5 /. subPars;
Res3k + (I Pi^2) Res4kD
,Print["Warning: Tensor integrals of rank higher than 3 only implemented for 4 denominators. Use other method."]
]  
]
]
(*---------------------------------------------------------------------*)


(*---------------------------------------------------------------------*)
(* OneLoop4kMod: Function to calculate OneLoop Integrals with 4 momenta and integrates with OneLopMod *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Calculate the One Loop integral for the full expression. Extends the OneLoop function of FC to the case 
              of 4 momenta in the numerator.
*)
  
OneLoop4kMod=Function[{k,amp},Module[{aux1,aux2,amp3k,amp4k,aux3,aux4,aux5,Res3k,Res4kD,subPars},
aux1=Calc[amp]//FCI;
aux2=aux1 /. Momentum[k, D] -> x Momentum[k, D];
amp4k=Coefficient[aux2, x, 4] /. x -> 1 // FCE;			  
amp3k=Calc[(aux1 - amp4k)]//Simplify;
GetPars[k,amp4k,False];
If[Error=="False",
If[ToString[amp3k] != "0", Res3k=OneLoopMod[k,amp3k],Res3k=0;];
Num4k=GetNumerator[amp4k];
subPars={Momentum[r1,D] -> Momentum[rr1,D], Momentum[r2,D]->Momentum[rr2,D],Momentum[r3,D]->Momentum[rr3,D] ,m0->mm0,m1->mm1,m2->mm2,m3->mm3} ;
aux3=IndividualFVkD[k,Num4k] ;
aux4 = 
  derD[derD[
      derD[derD[aux3, Pair[LorentzIndex[j1, D], Momentum[k, D]]], 
      Pair[LorentzIndex[j2, D], Momentum[k, D]]], 
      Pair[LorentzIndex[j3, D], Momentum[k, D]]], 
      Pair[LorentzIndex[j4, D], Momentum[k, D]]]/4! // Expand;
aux5 = Calc[aux4 Dabcd[j1, j2, j3, j4, LL1, LL2]];
Res4kD=aux5 /. subPars;
Res3k + (I Pi^2) Res4kD				    
,Print["Warning: Tensor integrals of rank higher than 3 only implemented for 4 denominators. Use other method."]
]
]
]
(*---------------------------------------------------------------------*)

  
(*---------------------------------------------------------------------*)				 
(* MyOneLoop: Function that integrates OneLoop with OneLoop4k *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Calculate the One Loop integral for the full expression. 
   Extends the OneLoop function of FC to the case  of 4 momenta 
   in the numerator.
*)

	
MyOneLoop=Function[{k,amp},Module[{aux1,aux2,amp4k,Loop4k,ResLoop},
Clear[x];				  
aux1=Calc[amp]//FCI;
aux2=aux1 /. Momentum[k, D] -> x Momentum[k, D];
amp4k=Coefficient[aux2, x, 4] /. x -> 1 // FCE;
If[ToString[amp4k] != "0",Loop4k=1,Loop4k=0];
If[Loop4k==1,ResLoop=OneLoop4k[k,amp];,ResLoop=OneLoop[k,amp]];
ResLoop
]
]
(*---------------------------------------------------------------------*)

  
(*---------------------------------------------------------------------*)
(* MyOneLoopMod: Function that integrates OneLoopMod with OneLoop4kMod *)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Calculate the One Loop integral for the full expression. 
              Uses OneLoopMod for the part with 3 momenta and
              OneLoop4kMod for the 4 momenta case.
*)
		 
MyOneLoopMod=Function[{k,amp},Module[{aux1,aux2,amp4k,ResLoop,Loop4k},
aux1=Calc[amp]//FCI;
If[ToString[aux1] == "0",ResLoop=0];
If[ToString[aux1] != "0",
        aux2=aux1 /. Momentum[k, D] -> x Momentum[k, D];
	amp8k=Coefficient[aux2, x, 8] /. x -> 1 // FCE;
	amp7k=Coefficient[aux2, x, 7] /. x -> 1 // FCE;
	amp6k=Coefficient[aux2, x, 6] /. x -> 1 // FCE;
	amp5k=Coefficient[aux2, x, 5] /. x -> 1 // FCE;
        amp4k=Coefficient[aux2, x, 4] /. x -> 1 // FCE;];
If[ToString[amp8k] != "0"||  ToString[amp7k] != "0"
   || ToString[amp6k] != "0" || ToString[amp5k] != "0",Print["Warning: Tensor integrals of rank higher than 3 only implemented for 4 denominators. Use other method."],   
   If[ToString[amp4k] != "0",Loop4k=1,Loop4k=0];
    If[Loop4k==1,ResLoop=OneLoop4kMod[k,amp];,ResLoop=OneLoopMod[k,amp]];
    ResLoop]
]
]
(*---------------------------------------------------------------------*)
  
(*---------------------------------------------------------------------*)
(* OneLoopMod: Function to calculate OneLoop Integrals up to 3 momenta.
   Method:
   1) Expand the expression in monomials. 
   2) Use the function OneLoopModMon for each monomial
   3) Sum everything 
*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Calculate the One Loop integral for the full expression. 
              It is limited to 3 momenta because of the limitation
              in the FC function OneLoop that it uses.
*)


OneLoopMod=Function[{k,amp},Module[{aux0,aux1,aux2},
  aux0=Expand[amp];
  If[Head[aux0]===Plus,
  aux1=Table[OneLoopModMon[k,aux0[[i]]],{i,1,Length[aux0]}];
  aux2=Sum[aux1[[i]],{i,1,Length[aux0]}];];
  If[Head[aux0]===Times,aux2=OneLoopModMon[k,aux0];];
  Res=aux2]]
(*---------------------------------------------------------------------*)      

  
(*---------------------------------------------------------------------*)
(* OneLoopModMon: Function to calculate OneLoop Integrals up to 3 momenta.
                  Might exist problems for expressions with both terms with 
                  and without epsilons it should be used inside OneLoopMod 
                  for each monomial.
*)
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Calculate the One Loop integral for each monomial of the 
   expression. It is limited to 3 momenta because of the limitation 
   in the FC function OneLoop that it uses.
*)


  
OneLoopModMon=Function[{k,amp},Module[{test,aux0,aux1,aux2,aux3,aux4a,aux4b,aux5,aux6,aux7,aux8,aux9,aux10,aux11,subToD,Num0k,Num1k,Num2k,Num3k,Num4k,DenAmp,NumAmp,Res0kD,Res1kD,Res3kD,Res},
				   
subToD = {LorentzIndex[iLo_] -> LorentzIndex[iLo, D], 
	  Momentum[xmom_] -> Momentum[xmom, D]};

test=Calc[amp];

If[ToString[test]!="0",                                   
DenAmp=GetDenominator[amp];
NumAmp=GetNumerator[amp];				   
aux0=Calc[NumAmp]//FCI;
aux1=aux0 /. {Momentum[k] -> xId Momentum[k],Momentum[k, D] -> xId Momentum[k, D]};
aux2=Calc[aux1];				   
Num4k=Coefficient[aux2, xId, 4] /. xId -> 1 // FCI//Expand;  
Num3k=Coefficient[aux2, xId, 3] /. xId -> 1 // FCI//Expand;
Num2k=Coefficient[aux2, xId, 2] /. xId -> 1 // FCI//Expand;
Num1k=Coefficient[aux2, xId, 1] /. xId -> 1 // FCI//Expand;
Num0k=Coefficient[aux2, xId, 0] /. xId -> 1 // FCI//Expand;
, Num0k=0;Num1k=0;Num2k=0;Num3k=0;Num4k=0;];

Res0kD=0;
If[ToString[Num0k]!="0", 
Res0kD=(Num0k OneLoop[k,DenAmp]//FCI)/. subToD;];
   
Res1kD=0;
If[ToString[Num1k]!="0",		   
aux3=IndividualFVkD[k,Num1k] ;
aux4a = 
derD[aux3, Pair[LorentzIndex[j1, D], Momentum[k, D]]]/1! // Expand;
aux4b=(OneLoop[k,FVD[k,j1] DenAmp]//FCI)/.subToD;
aux5 = Calc[aux4a aux4b];
Res1kD=aux5;
];
   
Res2kD=0;
If[ToString[Num2k]!="0",
aux6=IndividualFVkD[k,Num2k] ;
aux7a = derD[
       derD[aux6, Pair[LorentzIndex[j1, D], Momentum[k, D]]],
  Pair[LorentzIndex[j2, D], Momentum[k, D]]]/2! // Expand;
aux7b = (OneLoop[k,FVD[k,j1] FVD[k,j2] DenAmp]//FCI)/. subToD;
aux8 = Calc[aux7a aux7b];
Res2kD=aux8;
];

Res3kD=0;
If[ToString[Num3k]!="0",
aux9=IndividualFVkD[k,Num3k] ;
aux10a = derD[derD[
  derD[aux9, Pair[LorentzIndex[j1, D], Momentum[k, D]]],
  Pair[LorentzIndex[j2, D], Momentum[k, D]]],
  Pair[LorentzIndex[j3, D], Momentum[k, D]]]/3! // Expand;
aux10b = (OneLoop[k,FVD[k,j1] FVD[k,j2] FVD[k,j3] DenAmp]//FCI)/.subToD;
aux11 = Calc[aux10a aux10b];   
Res3kD=aux11;
   ];		   
Res=Res0kD + Res1kD + Res2kD + Res3kD			    
]
]
(*---------------------------------------------------------------------*)

  
(*---------------------------------------------------------------------*)
(* GetDenominator: Get denominator of the Loop integral *) 
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Get the Denominator of the Feynman Integral. Has problems if there are both epsilon terms and terms without
              epsilons. Therefore when used with OneLoopMod one has to separate the integral in monomials.
*)
   
GetDenominator=Function[{xamp},Module[{aux1,aux2,aux3,aux4,aux5,aux6,posi,posf,icount},
aux1=((Calc[xamp]//FeynAmpDenominatorCombine)//Simplify)//FCE;
aux2=InputForm[aux1];			       
aux3=ToString[aux2];
aux4=StringPosition[aux3, "FAD["];
aux5=StringPosition[aux3,"]"];
posi=aux4[[1]][[1]];
icount=0;
Do[If[aux5[[j]][[1]] > posi, icount = icount + 1; 
If[icount == 1, posf = aux5[[j]][[1]]]], {j, 1, Length[aux5]}];
aux6=StringTake[aux3, {posi,posf}];
ToExpression[aux6]				    
]];
(*---------------------------------------------------------------------*) 

  
(*---------------------------------------------------------------------*) 				      
(* GetNumerator: Get numerator of the Loop integral *) 
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Get the Numerator of the Feynman Integral. Has problems if there are both epsilon terms and terms without
              epsilons. Therefore when used with OneLoopMod one has to separate the integral in monomials.
*)

  
GetNumerator=Function[{xamp},Module[{aux0,aux1,aux2,aux3,aux4,aux5,aux6,LS,posi,posf,case,ix},
aux0=((Calc[xamp]//FeynAmpDenominatorCombine)//Simplify)//FCE;
aux1=InputForm[aux0];			       
aux2=ToString[aux1];
LS=StringLength[aux2];				    
aux3=StringPosition[aux2, {"FAD[", "]"}];


ix = 1;
While[Length[aux3] - ix >= 0, 
  If[StringTake[aux2, {aux3[[ix]][[1]], aux3[[ix]][[2]]}] == "FAD[", 
   posi = aux3[[ix]][[1]]; posf = aux3[[ix + 1]][[1]]]; ix++];  

				    
If[posi == 1 && posf == LS,case=1];
If[posi == 1 && posf < LS,case=2];
If[posi > 1 && posf == LS,case=3];
If[posi > 1 && posf < LS,case=4];

If[case==1, aux6="1";];

If[case==2, aux6=StringTake[aux2, {posf+2,LS}]];

If[case==3, aux6=StringTake[aux2, {1, posi-2}]];
				    
If[case==4, aux4=StringTake[aux2, {1, posi-1}];
aux5=StringTake[aux2, {posf+1,LS}];
aux6=StringJoin[aux4," 1 ",aux5];];

ToExpression[aux6]				    
 ]];
(*---------------------------------------------------------------------*) 


(*---------------------------------------------------------------------*)               
(* TCALoopExp: Replace four-vectors (of loop momentum) with expressions from Jorge's TCA appendix *) 
(* Created by: Duarte Fontes
  Email: duartefontes@tecnico.ulisboa.pt 
  Last Update: 20.12.2019
  Notes:
  1) this function has 6 arguments: {exp, Denos, Powk, k1, P, C}, where:
        -> exp is the expression to be acted on,
        -> Denos is the number of denominators,
        -> Powk is the power of loop momenta in the numerator of the loop integral,
        -> k1 is the loop momentum,
        -> P and C are the P and C variables from Jorge's TCA appendix;
  2) I am not including the factor I/(16 Pi^2); it should be included by hand;
  3) I am already including the Gamma factors; for example, with 3 denominators, I am already including the Gamma[3]=2;
  4) the expressions are written for D-dimensional objects; hence, we have MTD[a,b], and not MT[a,b], for example.
*)  
TCALoopExp=Function[{exp, Denos, Powk, k1, P, C}, Module[{X0,G,EP,mu,nu,a,b,d,myrep},
EP=\[CapitalDelta]\[Epsilon];
(* One denominator *)
If[Denos==1,
G=1;
If[Powk==0, myrep = d_ -> d*G*C*(1 + EP - Log[C]),
If[Powk==1, myrep = FVD[k1,mu_] -> 0,
If[Powk==2, myrep = FVD[k1,mu_] FVD[k1,nu_] -> G*(1/8) C^2 MTD[mu,nu](3 + 2*EP - 2*Log[C])]]],
(* Two denominators *)
If[Denos==2,
G=1;
If[Powk==0, myrep = d_ -> d*G*(EP - Log[C]),
If[Powk==1, myrep = FVD[k1,mu_] -> G*(-EP + Log[C]) FVD[P,mu],
If[Powk==2, myrep = FVD[k1,mu_] FVD[k1,nu_] -> G*(1/2)*(C MTD[mu,nu] (1 + EP - Log[C]) + 2*(EP - Log[C]) FVD[mu] FVD[nu]),
If[Powk==3, myrep = FVD[k1,mu_] FVD[k1,nu_] FVD[k1,a_] -> G*(1/2) (-C MTD[mu,nu] (1 + EP - Log[C]) FVD[P,a] - C MTD[nu,a] * (1
                      + EP - Log[C]) FVD[P,mu] - C MTD[mu,a] (1 + EP - Log[C]) FVD[P,nu] - 2 FVD[P,a] FVD[P,mu] FVD[P,nu] (EP - Log[C]))]]]],
(* Three denominators *)
If[Denos==3,
G=2;
If[Powk==0, myrep = d_ -> d*G*(-1)/(2*C),
If[Powk==1, myrep = FVD[k1,mu_] -> G*FVD[P, mu]/(2*C),
If[Powk==2, myrep = FVD[k1,mu_] FVD[k1,nu_] -> G*1/(4*C)*(C*MTD[mu,nu] (EP - Log[C]) - 2*FVD[P,mu] FVD[P,nu]),
If[Powk==3, myrep = FVD[k1,mu_] FVD[k1,nu_] FVD[k1,a_] -> G*1/(4*C)*(C*MTD[mu,nu] (-EP + Log[C]) FVD[P,a] + C*MTD[nu,a] (-EP + Log[C]) FVD[P,mu]
                                                                                        + C*MTD[mu,a] (-EP + Log[C]) FVD[P,nu] + 2 FVD[P,a] FVD[P,mu] FVD[P,nu]),
If[Powk==4, myrep = FVD[k1,mu_] FVD[k1,nu_] FVD[k1,a_] FVD[k1,b_] -> G*1/(8*C)*(C^2 (1 + EP - Log[C]) (MTD[mu,a] MTD[nu,b] + MTD[mu,b] MTD[nu,a]
                                                                                                                + MTD[a,b] MTD[mu,nu])
                                                        + 2*C (EP - Log[C])*(MTD[mu,nu] FVD[P,a] FVD[P,b] + MTD[nu,b] FVD[P,a] FVD[P,mu] + MTD[nu,a] FVD[P,b] FVD[P,mu]
                                                                                               + MTD[mu,a] FVD[P,b] FVD[P,nu] + MTD[mu,b] FVD[P,a] FVD[P,nu] + MTD[a,b] FVD[P,mu] FVD[P,nu])
                                                        - 4 FVD[P,a] FVD[P,b] FVD[P,mu] FVD[P,nu])]]]]],
(* Four denominators *)
If[Denos==4,
G=6;
If[Powk==0, myrep = d_ -> d*G*1/(6*C^2),
If[Powk==1, myrep = FVD[k1,mu_] -> G*(-1)/(6*C^2) FVD[P,mu],
If[Powk==2, myrep = FVD[k1,mu_] FVD[k1,nu_] -> G*(-1)/(12*C^2) (C MTD[mu,nu] - 2 FVD[P,mu] FVD[P,nu]),
If[Powk==3, myrep = FVD[k1,mu_] FVD[k1,nu_] FVD[k1,a_] -> G*1/(12*C^2) (C*(MTD[mu,nu] FVD[P,a] + MTD[nu,a] FVD[P,mu] + MTD[mu,a] FVD[P,nu])
                                                                                      - 2 FVD[P,a] FVD[P,mu] FVD[P,nu]),
If[Powk==4, myrep = FVD[k1,mu_] FVD[k1,nu_] FVD[k1,a_] FVD[k1,b_] -> G*1/(24*C^2) (C^2 (EP - Log[C]) (MTD[mu,a] MTD[nu,b] + MTD[mu,b] MTD[nu,a]
                                                                                                                + MTD[a,b] MTD[mu,nu])
                                                  -2*C*(MTD[mu,nu] FVD[P,a] FVD[P,b] + MTD[nu,b] FVD[P,a] FVD[P,mu] + MTD[nu,a] FVD[P,b] FVD[P,mu]
                                                                                            + MTD[mu,a] FVD[P,b] FVD[P,nu] + MTD[mu,b] FVD[P,a] FVD[P,nu] + MTD[a,b] FVD[P,mu] FVD[P,nu])
                                                  + 4 FVD[P,a] FVD[P,b] FVD[P,mu] FVD[P,nu])]]]]]]]]];
X0=exp/.myrep]];
(*---------------------------------------------------------------------*) 



(*---------------------------------------------------------------------*) 
(* Change expression to D Dimensions *) 
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Changes all the Lorentz quantitues to Dimension D
*)

ChangeToD=Function[{exp},Module[{aux1,aux2,aux3,aux4,subToD,pp,qq,rr,p1,p2,iL1,iL2,iL3,iL4,
                                                                            zz1, zz2, zz3, zz4, zz5, zz6, zz7, zz8, zz9, zz10, zz11, zz12, zz13, zz14, zz15, zz16, zz17, zz18, zz19,  zz20, zz21},
aux1=exp//FCE;
subToD ={FV[pp_, iL1_] -> FVD[pp, iL1], SP[qq_, rr_] -> SPD[qq, rr], GS[p1_,p2_] -> GSD[p1,p2], MT[iL2_, iL3_] -> MTD[iL2, iL3]};
aux2=(aux1 /. subToD) // FCI;
aux3=aux2/.{DiracGamma[Momentum[p1_]] -> DiracGamma[Momentum[p1, D], D],
            DiracGamma[LorentzIndex[iL4_]] -> DiracGamma[LorentzIndex[iL4,D],D]};  
aux4=(aux3 // FCE) /. {
FAD[zz21_] -> FAD[zz21, Dimension -> D],
FAD[{zz1_, zz2_}] -> FAD[{zz1, zz2}, Dimension -> D], 
FAD[{zz3_, zz4_}, {zz5_, zz6_}] -> FAD[{zz3, zz4}, {zz5, zz6}, Dimension -> D],
FAD[{zz7_, zz8_}, {zz9_, zz10_}, {zz11_, zz12_}] -> FAD[{zz7, zz8}, {zz9, zz10}, {zz11, zz12}, Dimension -> D],
FAD[{zz13_, zz14_}, {zz15_, zz16_}, {zz17_, zz18_}, {zz19_, zz20_}] -> FAD[{zz13, zz14}, {zz15, zz16}, {zz17, zz18}, {zz19, zz20}, Dimension -> D]}]];
(*---------------------------------------------------------------------*) 


(*---------------------------------------------------------------------*)   
(* Change expression to 4 Dimensions *) 
(* Created by: Jorge C. Romao
   Email: jorge.romao@tecnico.ulisboa.pt 
   Last Update: 01.08.2019

   Objective: Changes all the Lorentz quantitues to D=4
*)
  
ChangeTo4 = Function[{exp}, Module[{aux1, aux2, aux3, zz1, zz2, zz3, zz4, zz5, zz6, zz7, zz8, zz9, zz10, zz11, zz12, zz13, zz14, zz15, zz16, zz17, zz18, zz19,  zz20, zz21},
aux1 = ChangeToD[exp] // FCI;
aux2 = (aux1 /. D -> 4) // FCE;
aux3 =  aux2 /. {
FAD[zz21_, Dimension -> 4] -> FAD[zz21],
FAD[{zz1_, zz2_}, Dimension -> 4] -> FAD[{zz1, zz2}], 
FAD[{zz3_, zz4_}, {zz5_, zz6_}, Dimension -> 4] -> FAD[{zz3, zz4}, {zz5, zz6}],
FAD[{zz7_, zz8_}, {zz9_, zz10_}, {zz11_, zz12_}, Dimension -> 4] -> FAD[{zz7, zz8}, {zz9, zz10}, {zz11, zz12}],
FAD[{zz13_, zz14_}, {zz15_, zz16_}, {zz17_, zz18_}, {zz19_, zz20_}, Dimension -> 4] -> FAD[{zz13, zz14}, {zz15, zz16}, {zz17, zz18}, {zz19, zz20}]}]];
(*---------------------------------------------------------------------*) 



FermionLineNoG5 = 
 Function[line,Module[{aux1,aux2,yy},
    aux1 = DiracSimplify[line] // FCE; 
   aux2 = DiracSimplify[aux1 /. GA[5] -> yy GA[5]]; 
   Coefficient[aux2, yy, 0]]];

FermionLineG5 = 
  Function[line,Module[{aux1,aux2,yy},
  aux1 = DiracSimplify[line] // FCE; 
   aux2 = DiracSimplify[aux1 /. GA[5] -> yy GA[5]]; 
   Coefficient[aux2, yy, 1]]];

FermionLineG5Part = 
  Function[line, Module[{aux},
  aux = FermionLineG5[line] // FCE; 
  DiracSimplify[aux.GA[5]]]];

TrG5=Function[exp,Module[{aux,auxNoG5,auxG5,fcng5},
auxNoG5=Tr[FermionLineNoG5[exp]];
auxG5=Tr[ChangeTo4[FermionLineG5[exp]]];
aux=auxNoG5 + auxG5;   
fcng5=aux /. {Tr[0] -> 0}]];
  


	
